/* SPDX-FileCopyrightText: 2016 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup draw_engine
 */

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

extern DrawEngineType draw_engine_basic_type;

#ifdef __cplusplus
}
#endif
