/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup pythonintern
 */

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

extern PyMethodDef BPY_rna_uilayout_introspect_method_def;

#ifdef __cplusplus
}
#endif
