/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup DNA
 */

#pragma once

/* Struct members on own line. */
/* clang-format off */

/* -------------------------------------------------------------------- */
/** \name Collection Struct
 * \{ */

#define _DNA_DEFAULT_Collection \
  { \
    .color_tag = COLLECTION_COLOR_NONE, \
  }

/** \} */

/* clang-format on */
