# SPDX-FileCopyrightText: 2022-2023 Blender Foundation
#
# SPDX-License-Identifier: GPL-2.0-or-later

# ##### BEGIN AUTOGENERATED I18N SECTION #####
# NOTE: You can safely move around this auto-generated block (with the begin/end markers!),
#       and edit the translations by hand.
#       Just carefully respect the format of the tuple!

# Tuple of tuples:
# ((msgctxt, msgid), (sources, gen_comments), (lang, translation, (is_fuzzy, comments)), ...)
translations_tuple = (
    (("*", ""),
     ((), ()),
     ("fr_FR", "Project-Id-Version: AnimAll 0.10.0 (0)\n",
               (False,
                ("Blender's translation file (po format).",
                 "Copyright (C) 2022-2023 The Blender Foundation.",
                 "This file is distributed under the same license as the Blender package.",
                 "Damien Picard <dam.pic@free.fr>, 2022."))),
    ),
    (("*", "Tab Category"),
     (("bpy.types.AnimallAddonPreferences.category",),
      ()),
     ("fr_FR", "Catégorie d’onglet",
               (False, ())),
    ),
    (("*", "Choose a name for the category of the panel"),
     (("bpy.types.AnimallAddonPreferences.category",),
      ()),
     ("fr_FR", "Choisir un nom pour la catégorie du panneau",
               (False, ())),
    ),
    (("Operator", "Clear Animation"),
     (("bpy.types.ANIM_OT_clear_animation_animall",),
      ()),
     ("fr_FR", "Effacer l’animation",
               (False, ())),
    ),
    (("*", "Delete all keyframes for this object\nIf in a specific case it doesn't work\ntry to delete the keys manually"),
     (("bpy.types.ANIM_OT_clear_animation_animall",),
      ()),
     ("fr_FR", "Supprimer toutes les images clés pour cet objet.\n"
               "En cas d’échec, essayez de les supprimer manuellement",
               (False, ())),
    ),
    (("Operator", "Insert Key"),
     (("bpy.types.ANIM_OT_insert_keyframe_animall",),
      ()),
     ("fr_FR", "Insérer une clé",
               (False, ())),
    ),
    (("*", "Insert a Keyframe"),
     (("bpy.types.ANIM_OT_insert_keyframe_animall",),
      ()),
     ("fr_FR", "Insérer une image clé",
               (False, ())),
    ),
    (("Operator", "Delete Key"),
     (("bpy.types.ANIM_OT_delete_keyframe_animall",),
      ()),
     ("fr_FR", "Supprimer l’image clé",
               (False, ())),
    ),
    (("*", "Delete a Keyframe"),
     (("bpy.types.ANIM_OT_delete_keyframe_animall",),
      ()),
     ("fr_FR", "Supprimer une image clé",
               (False, ())),
    ),
    (("Operator", "Update Attribute Animation"),
     (("bpy.types.ANIM_OT_update_attribute_animation_animall",),
      ()),
     ("fr_FR", "Mettre à jour l’animation des attributs",
               (False, ())),
    ),
    (("*", "Update attributes from the old format"),
     (("bpy.types.ANIM_OT_update_attribute_animation_animall",),
      ()),
     ("fr_FR", "Mettre à jour les attributs depuis l’ancien format",
               (False, ())),
    ),
    (("*", "Animate"),
     (("bpy.types.VIEW3D_PT_animall",),
      ()),
     ("fr_FR", "Animer",
               (False, ())),
    ),
    (("*", "Insert keyframes on active attribute values"),
     (("bpy.types.AnimallProperties.key_active_attribute",),
      ()),
     ("fr_FR", "Insérer des clés sur l’attribut actif",
               (False, ())),
    ),
    (("*", "Insert keyframes on edge bevel weight"),
     (("bpy.types.AnimallProperties.key_edge_bevel",),
      ()),
     ("fr_FR", "Insérer des clés sur les poids de biseau d’arête",
               (False, ())),
    ),
    (("*", "Edge Crease"),
     (("bpy.types.AnimallProperties.key_edge_crease",),
      ()),
     ("fr_FR", "Plis d’arêtes",
               (False, ())),
    ),
    (("*", "Insert keyframes on edge creases"),
     (("bpy.types.AnimallProperties.key_edge_crease",),
      ()),
     ("fr_FR", "Insérer des clés sur les plis d’arête",
               (False, ())),
    ),
    (("*", "Insert keyframes on face material indices"),
     (("bpy.types.AnimallProperties.key_material_index",),
      ()),
     ("fr_FR", "Insérer des clés sur les indices de matériaux",
               (False, ())),
    ),
    (("*", "Insert keyframes on point locations"),
     (("bpy.types.AnimallProperties.key_point_location",),
      ()),
     ("fr_FR", "Insérer des clés sur les positions des points",
               (False, ())),
    ),
    (("*", "Insert keyframes on point radius (Shrink/Fatten)"),
     (("bpy.types.AnimallProperties.key_radius",),
      ()),
     ("fr_FR", "Insérer des clés sur le rayon de rayon de point (épaisseur de la courbe)",
               (False, ())),
    ),
    (("*", "Key Selected Only"),
     (("bpy.types.AnimallProperties.key_selected",),
      ()),
     ("fr_FR", "Sélection uniquement",
               (False, ())),
    ),
    (("*", "Insert keyframes only on selected elements"),
     (("bpy.types.AnimallProperties.key_selected",),
      ()),
     ("fr_FR", "Insérer des images clés seulement sur les éléments sélectionnés",
               (False, ())),
    ),
    (("*", "Insert keyframes on active Shape Key layer"),
     (("bpy.types.AnimallProperties.key_shape_key",),
      ()),
     ("fr_FR", "Insérer des clés sur le calque de clé de forme actif",
               (False, ())),
    ),
    (("*", "Insert keyframes on point tilt"),
     (("bpy.types.AnimallProperties.key_tilt",),
      ()),
     ("fr_FR", "Insérer des clés sur l’inclinaison des points",
               (False, ())),
    ),
    (("*", "Insert keyframes on active UV coordinates"),
     (("bpy.types.AnimallProperties.key_uvs",),
      ()),
     ("fr_FR", "Insérer des clés sur les coordonnées UV actives",
               (False, ())),
    ),
    (("*", "Insert keyframes on vertex bevel weight"),
     (("bpy.types.AnimallProperties.key_vertex_bevel",),
      ()),
     ("fr_FR", "Insérer des clés sur les poids de biseau des sommets",
               (False, ())),
    ),
    (("*", "Insert keyframes on vertex crease weight"),
     (("bpy.types.AnimallProperties.key_vertex_crease",),
      ()),
     ("fr_FR", "Insérer des clés sur les plis de sommets",
               (False, ())),
    ),
    (("*", "Insert keyframes on active vertex group values"),
     (("bpy.types.AnimallProperties.key_vertex_group",),
      ()),
     ("fr_FR", "Insérer des clés sur les valeurs des groupes de sommets",
               (False, ())),
    ),
    (("*", "AnimAll"),
     (("Add-on AnimAll info: name",),
      ()),
     ("fr_FR", "AnimAll",
               (False, ())),
    ),
    (("*", "Key:"),
     (("scripts/addons/animation_animall/__init__.py:200",),
      ()),
     ("fr_FR", "Insérer :",
               (False, ())),
    ),
    (("*", "Tab Category:"),
     (("scripts/addons/animation_animall/__init__.py:704",),
      ()),
     ("fr_FR", "Catégorie d’onglet :",
               (False, ())),
    ),
    (("*", "Points"),
     (("scripts/addons/animation_animall/__init__.py:206",
       "scripts/addons/animation_animall/__init__.py:213",
       "scripts/addons/animation_animall/__init__.py:243"),
      ()),
     ("fr_FR", "Points",
               (False, ())),
    ),
    (("*", "Others"),
     (("scripts/addons/animation_animall/__init__.py:209",
       "scripts/addons/animation_animall/__init__.py:226",
       "scripts/addons/animation_animall/__init__.py:251"),
      ()),
     ("fr_FR", "Autres",
               (False, ())),
    ),
    (("*", "Bevel"),
     (("scripts/addons/animation_animall/__init__.py:215",
       "scripts/addons/animation_animall/__init__.py:220"),
      ()),
     ("fr_FR", "Biseau",
               (False, ())),
    ),
    (("*", "Edges"),
     (("scripts/addons/animation_animall/__init__.py:219",),
      ()),
     ("fr_FR", "Arêtes",
               (False, ())),
    ),
    (("*", "Crease"),
     (("scripts/addons/animation_animall/__init__.py:216",
       "scripts/addons/animation_animall/__init__.py:221",),
      ()),
     ("fr_FR", "Plis",
               (False, ())),
    ),
    (("*", "Faces"),
     (("scripts/addons/animation_animall/__init__.py:223",),
      ()),
     ("fr_FR", "Faces",
               (False, ())),
    ),
    (("*", "\"Location\" and \"Shape Key\" are redundant?"),
     (("scripts/addons/animation_animall/__init__.py:273",),
      ()),
     ("fr_FR", "\"Position\" et \"Clé de forme\" sont redondants ?",
               (False, ())),
    ),
    (("*", "Splines"),
     (("scripts/addons/animation_animall/__init__.py:248",),
      ()),
     ("fr_FR", "Splines",
               (False, ())),
    ),
    (("*", "Maybe set \"%s\" to 1.0?"),
     (("scripts/addons/animation_animall/__init__.py:264"),
      ()),
     ("fr_FR", "Essayez de mettre « %s » à 1.0 ?",
               (False, ())),
    ),
    (("*", "Cannot key on Basis Shape"),
     (("scripts/addons/animation_animall/__init__.py:267",),
      ()),
     ("fr_FR", "Impossible d’ajouter une clé sur la forme de base",
               (False, ())),
    ),
    (("*", "No active Shape Key"),
     (("scripts/addons/animation_animall/__init__.py:270",),
      ()),
     ("fr_FR", "Pas de clé de forme active",
               (False, ())),
    ),
    (("*", "Clear Animation could not be performed"),
     (("scripts/addons/animation_animall/__init__.py:615",),
      ()),
     ("fr_FR", "La suppression de l’animation n’a pas pu aboutir",
               (False, ())),
    ),
    (("*", "Object includes old-style attributes. Consider updating them."),
     (("scripts/addons/animation_animall/__init__.py:237",),
      ()),
     ("fr_FR", "L’objet contient des attributs à l’ancien format. Veuillez les mettre à jour.",
               (False, ())),
    ),
    (("*", "Vertex %s"),
     (("scripts/addons/animation_animall/__init__.py:141",
       "scripts/addons/animation_animall/__init__.py:368",
       "scripts/addons/animation_animall/__init__.py:382",
       "scripts/addons/animation_animall/__init__.py:416"),
      ()),
     ("fr_FR", "Sommet %s",
               (False, ())),
    ),
    (("*", "Edge %s"),
     (("scripts/addons/animation_animall/__init__.py:143"),
      ()),
     ("fr_FR", "Arête %s",
               (False, ())),
    ),
    (("*", "Point %s"),
     (("scripts/addons/animation_animall/__init__.py:320",),
      ()),
     ("fr_FR", "Point %s",
               (False, ())),
    ),
    (("*", "Spline %s"),
     (("scripts/addons/animation_animall/__init__.py:328",),
      ()),
     ("fr_FR", "Spline %s",
               (False, ())),
    ),
    (("*", "Face %s"),
     (("scripts/addons/animation_animall/__init__.py:145",
       "scripts/addons/animation_animall/__init__.py:395"),
      ()),
     ("fr_FR", "Face %s",
               (False, ())),
    ),
    (("*", "%s Point %s"),
     (("scripts/addons/animation_animall/__init__.py:315",),
      ()),
     ("fr_FR", "%s Point %s",
               (False, ())),
    ),
    (("*", "Loop %s"),
     (("scripts/addons/animation_animall/__init__.py:147",),
      ()),
     ("fr_FR", "Boucle %s",
               (False, ())),
    ),
    (("*", "UV Layer %s"),
     (("scripts/addons/animation_animall/__init__.py:409",),
      ()),
     ("fr_FR", "Calque UV %s",
               (False, ())),
    ),
    (("*", "%s Vertex %s"),
     (("scripts/addons/animation_animall/__init__.py:416",),
      ()),
     ("fr_FR", "%s Sommet %s",
               (False, ())),
    ),
    (("*", "Spline %s CV %s"),
     (("scripts/addons/animation_animall/__init__.py:338",
       "scripts/addons/animation_animall/__init__.py:339",
       "scripts/addons/animation_animall/__init__.py:340",
       "scripts/addons/animation_animall/__init__.py:343",
       "scripts/addons/animation_animall/__init__.py:346",
       "scripts/addons/animation_animall/__init__.py:352",
       "scripts/addons/animation_animall/__init__.py:355",
       "scripts/addons/animation_animall/__init__.py:358"),
      ()),
     ("fr_FR", "Spline %s Point %s",
               (False, ())),
    ),
    (("*", "%s Spline %s CV %s"),
     (("scripts/addons/animation_animall/__init__.py:432",
       "scripts/addons/animation_animall/__init__.py:434",
       "scripts/addons/animation_animall/__init__.py:437",
       "scripts/addons/animation_animall/__init__.py:440",
       "scripts/addons/animation_animall/__init__.py:442",
       "scripts/addons/animation_animall/__init__.py:450",
       "scripts/addons/animation_animall/__init__.py:452",
       "scripts/addons/animation_animall/__init__.py:453"),
      ()),
     ("fr_FR", "%s Spline %s Point %s",
               (False, ())),
    ),
    (("*", "3D View > Toolbox > Animation tab > AnimAll"),
     (("Add-on AnimAll info: location",),
      ()),
     ("fr_FR", "Vue 3D > Panneau N > Onglet Animer > AnimAll",
               (False, ())),
    ),
    (("*", "Allows animation of mesh, lattice, curve and surface data"),
     (("Add-on AnimAll info: description",),
      ()),
     ("fr_FR", "Permet d’animer les données de maillages, de lattices, de courbes et de surfaces",
               (False, ())),
    ),
)

translations_dict = {}
for msg in translations_tuple:
    key = msg[0]
    for lang, trans, (is_fuzzy, comments) in msg[2:]:
        if trans and not is_fuzzy:
            translations_dict.setdefault(lang, {})[key] = trans

# ##### END AUTOGENERATED I18N SECTION #####
