/* SPDX-FileCopyrightText: 2023 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

#pragma once

struct ID;
struct SpaceSpreadsheet;

ID *ED_spreadsheet_get_current_id(const SpaceSpreadsheet *sspreadsheet);
