/* SPDX-FileCopyrightText: 2012 Blender Authors
 *
 * SPDX-License-Identifier: GPL-2.0-or-later */

/** \file
 * \ingroup pybmesh
 */

#pragma once

#ifdef __cplusplus
extern "C" {
#endif

PyObject *BPyInit_bmesh_utils(void);

#ifdef __cplusplus
}
#endif
