/* SPDX-FileCopyrightText: 2009 Benoit Bolsee
 *
 * SPDX-License-Identifier: LGPL-2.1-or-later */

/** \file
 * \ingroup intern_itasc
 */

#include "eigen_types.hpp"

const KDL::Frame iTaSC::F_identity(Rotation::Identity(),Vector::Zero());


